
import matplotlib.pyplot as plt
import matplotlib.colors as colors
import numpy as np
import scipy.ndimage as spImg
import math



class Region(object):
       
    def __init__(self, label, cx, cy, xmin, xmax, ymin, ymax, count):
        self.label = label
        self.cx = cx
        self.cy = cy
        self.xmin = xmin
        self.xmax = xmax
        self.ymin = ymin
        self.ymax = ymax
        self.count = count
        
        self.pixels = []
        self.M = np.zeros((2, 2))
        self.eigenvectors = None
        self.eigenvalues = None
        
        
        
    def computeEigenvectorsAndValues(self):
        
        fxxSum = 0
        fyySum = 0
        fxySum = 0
        for p in self.pixels:
            fxxSum += (p[0] - self.cx) ** 2
            fyySum += (p[1] - self.cy) ** 2
            fxySum += (p[0] - self.cx) * (p[1] - self.cy)
        fxxSum /= self.count
        fyySum /= self.count
        fxySum /= self.count
        
        self.M[0, 0] = fxxSum
        self.M[0, 1] = fxySum
        self.M[1, 0] = fxySum
        self.M[1, 1] = fyySum
        
        e, v = np.linalg.eig(self.M)
        self.eigenvalues = e
        self.eigenvectors = v
        
        
        
    def draw(self):
              
        # Hauptachsen zeichnen:
        d1 = 2 * np.sqrt(self.eigenvalues[0])
        h1x = (self.cx, self.cx + self.eigenvectors[0, 0] * d1)
        h1y = (self.cy, self.cy + self.eigenvectors[1, 0] * d1)
        
        d2 = 2 * np.sqrt(self.eigenvalues[1])
        h2x = (self.cx, self.cx + self.eigenvectors[0, 1] * d2)
        h2y = (self.cy, self.cy + self.eigenvectors[1, 1] * d2)
        
        c1 = "red"
        c2 = "orange"
        if(self.eigenvalues[0] < self.eigenvalues[1]):
            c1 = "orange"
            c2 = "red"
           
        plt.plot(h1x, h1y, color=c1)
        plt.plot(h2x, h2y, color=c2)
        
        # Bounding-Box zeichnen:
        r = plt.Rectangle((self.xmin-1, self.ymin-1), self.xmax - self.xmin + 2, self.ymax - self.ymin + 2, color="blue", fill=False)
        plt.gca().add_artist(r)
        
        # Label zeichnen:
        plt.text(self.cx, self.cy, str(int(self.label)), color="red")
        
    


def normalize(image):
    
    minVal = np.min(image)
    maxVal = np.max(image)
    d = maxVal - minVal
    
    result = (image - minVal) / d
    
    return result



def preprocessImage(image):
    
    # Maximumsfilter anwenden:
    shadingImg = spImg.filters.maximum_filter(image, (11, 11))
    
    # Gaussfilter anwenden:
    shadingImg = spImg.filters.gaussian_filter(shadingImg, 35)
    
    # Beleuchtungsunterschiede entfernen:    
    epsilon = 0.1 # Um Division durch 0 zu vermeiden.
    result = image / (shadingImg + epsilon)
    
    # Medianfilter anwenden:
    result = spImg.filters.median_filter(result, (4, 4))
    
    print "Preprocessing done."
    
    return normalize(result)




def otsuSeparation(image):
    
    h, w = image.shape
    
    # normalisiertes Histogamm erstellen:
    H = np.zeros(256)
    for y in xrange(h):
        for x in xrange(w):
            value = int(image[y, x] * 255)
            H[value] += 1
    H = H / (h * w)
      
      
    # Durchschnittlicher globaler Grauwert des Bildes:
    mG = np.mean(image)
        
    # k optimieren:
    bestKQuality = -1
    bestK = -1
    for k in xrange(len(H)):
                
        # Klassenwahrscheinlichkeiten:
        P1k = np.sum(H[:k])
        P2k = np.sum(H[k:])
        
        # Durchschnittliche Grauwerte:
        sum1 = 0
        for i in xrange(k): sum1 += i * H[i]
        m1k = sum1 / P1k if P1k > 0 else 0
        sum2 = 0
        for i in xrange(k, 256): sum2 += i * H[i]
        m2k = sum2 / P2k if P2k > 0 else 0
        
        # Varianz zwischen den Klassen:
        sigmaB2k = P1k * (m1k - mG) ** 2 + P2k * (m2k - mG) ** 2
        
        # Guete des Schwellenwerts k:
        kQuality = sigmaB2k
        
        # Bestes k updaten / finden:
        if kQuality > bestKQuality: 
            bestK = k
            bestKQuality = kQuality
       
    # Bild mit bestem k segmentieren:
    border = bestK / 255.0
    result = np.array([[ (1.0 if v > border else 0.0) for v in row] for row in image])
    print "Best k:", bestK
    return result



# Rekursive Variante (siehe Skript). 
# Problem: Grosse Regionen brauchen u. U. zu viele Rekursionen.
# Der Fall tritt mit den Reisbildern aber nicht auf, 
# wenn man den Hintergrund gesondert behandelt.
def floodFillRecursive(image, labelMap, seedValue, x, y, label):
    
    h, w = image.shape
    
    if(image[y, x] == seedValue):
        labelMap[y, x] = label # Label vergeben
        # Nachbarn besuchen:
        if(x - 1 >= 0 and labelMap[y, x - 1] == -1):
            floodFillRecursive(image, labelMap, seedValue, x - 1, y, label)
        if(x + 1 < w and labelMap[y, x + 1] == -1):
            floodFillRecursive(image, labelMap, seedValue, x + 1, y, label)
        if(y - 1 >= 0 and labelMap[y - 1, x] == -1):
            floodFillRecursive(image, labelMap, seedValue, x, y - 1, label)
        if(y + 1 < h and labelMap[y + 1, x] == -1):
            floodFillRecursive(image, labelMap, seedValue, x, y + 1, label)
            
            
            
# Iterative Variante, keine Probleme mit grossen Regionen.          
def floodFillIterative(image, labelMap, seedValue, x, y, label):
    
    h, w = image.shape
    
    queue = [] # Queue fuer noch zu betrachtende Pixel
    
    labelMap[y, x] = label # Seed-Pixel labeln
    queue.append((x, y)) # Seed-Pixel in die Queue stellen
    
    # Solange Queue nicht leer:
    while(not len(queue) == 0):
        
        topX, topY = queue.pop(0) # Naechstes Pixel aus der Queue holen.
        
        # Nachbarn pruefen, ggf. labeln und in die Queue stellen:
        if(topX - 1 >= 0 and image[topY, topX - 1] == seedValue and labelMap[topY, topX - 1] == -1):
            labelMap[topY, topX - 1] = label
            queue.append((topX - 1, topY))
        if(topX + 1 < w and image[topY, topX + 1] == seedValue and labelMap[topY, topX + 1] == -1):
            labelMap[topY, topX + 1] = label
            queue.append((topX + 1, topY))
        if(topY - 1 >= 0 and image[topY - 1, topX] == seedValue and labelMap[topY - 1, topX] == -1):
            labelMap[topY - 1, topX] = label
            queue.append((topX, topY - 1))
        if(topY + 1 < h and image[topY + 1, topX] == seedValue and labelMap[topY + 1, topX] == -1):
            labelMap[topY + 1, topX] = label
            queue.append((topX, topY + 1))
        
    

def regionLabeling(image):
    
    h, w = image.shape   
     
    labelMap = np.ones((h, w)) * -1
    label = 0
    
    for y in xrange(h):
        for x in xrange(w):          
            if labelMap[y, x] == -1: # Noch kein Label...
                if(image[y, x] == 0): # Pixel gehoert zum Hintergrund (= Schwarz)
                    labelMap[y, x] = 0
                else:
                    label += 1   
                    floodFillIterative(image, labelMap, image[y, x], x, y, label)
                
    print "There are", label, "different labels."
    
    return labelMap



def createRegions(labelMap):
    
    h, w = labelMap.shape
    
    regionMap = {}
    
    for y in xrange(h):
        print "Working at labelMap row", y
        for x in xrange(w):  
            label = labelMap[y, x] # Label lesen
            
            if(not regionMap.has_key(label)): # Wenn noch keine Region erstellt fuer dieses Label:
                
                region = Region(label, x, y, x, x, y, y, 1)
                region.pixels.append((x, y))
                regionMap.setdefault(label, region)
            else: # Wenn Region schon vorhanden:
                region = regionMap.get(label) # Region holen
                # Werte anpassen:
                region.cx += x
                region.cy += y
                region.xmin = min(region.xmin, x)
                region.xmax = max(region.xmax, x)
                region.ymin = min(region.ymin, y)
                region.ymax = max(region.ymax, y)
                region.count += 1
                region.pixels.append((x, y))
    
    
    # Positionen mitteln:
    for region in regionMap.itervalues():
        region.cx /= region.count
        region.cy /= region.count
        
    # Eigenwerte bestimmen:
    for region in regionMap.itervalues():
        region.computeEigenvectorsAndValues()
          
        
    return regionMap
            



def markRegions(regionMap):
    
    for region in regionMap.itervalues():
                       
        if(region.label == 0):
            #plt.text(10, 30, str("Background"), color="blue")
            pass
        else:
            region.draw()
        

        


def aufgabe2():
    
    input = plt.imread('reis1.png')
    preprocessed = preprocessImage(input)
    resultOtsu = otsuSeparation(preprocessed)
    
    labelMap = regionLabeling(resultOtsu) 
    regionMap = createRegions(labelMap)
    

    plt.figure(1)
    #plt.subplot(121)
    #plt.imshow(input, cmap=plt.cm.gray, norm=colors.NoNorm())
    plt.subplot(111)
    plt.imshow(resultOtsu, cmap=plt.cm.gray, norm=colors.NoNorm())
    markRegions(regionMap)
    plt.show()
    
    
aufgabe2()
